"""
MIT License

Copyright (c) 2019 Luke Park

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from known_side_effects.argument_matcher import called_arguments_match, when_arguments_match
from known_side_effects.exceptions import UnmatchedArguments


class SideEffectFactory:
    def __init__(self):
        self._whens = []
        self._otherwise = None
        self._always = None

    def when(self, *arguments, **kwargs):
        # Attempt to match arguments and return that when rather
        # creating a new one
        for when in self._whens:
            if when_arguments_match(when, arguments, kwargs):
                return when

        when = When(self, arguments, kwargs)
        self._whens.append(when)
        return when

    def return_or_raise(self, response):
        should_raise, response = response
        if should_raise:
            raise response
        return response

    def __call__(self, *args, **kwargs):
        """
        side_effect
        """
        if self._always:
            return self.return_or_raise(self._always)
        for when in self._whens:
            if called_arguments_match(when, args, kwargs):
                response = when.get_response()
                return self.return_or_raise(response)

        if self._otherwise:
            return self.return_or_raise(self._otherwise)

        raise UnmatchedArguments(*args, **kwargs)


class When:
    def __init__(self, parent_side_effect, expected_arguments, expected_kwargs):
        self.expected_arguments = expected_arguments
        self.expected_kwargs = expected_kwargs
        self.parent_side_effect = parent_side_effect
        self._responses = []

    def then(self, response):
        self._responses.append((False, response))
        return self

    def then_raise(self, response):
        self._responses.append((True, response))
        return self

    def get_response(self):
        response = self._responses[0]
        if len(self._responses) > 1:
            response = self._responses.pop(0)
        return response

    def otherwise(self, response):
        self.parent_side_effect._otherwise = (False, response)

    def otherwise_raise(self, response):
        self.parent_side_effect._otherwise = (True, response)

    def always(self, response):
        self.parent_side_effect._always = (False, response)

    def always_raise(self, response):
        self.parent_side_effect._always = (True, response)



"""
MIT License

Copyright (c) 2019 Luke Park

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import List, Any, Dict

from hamcrest.core.matcher import Matcher

__all__ = [
    'match_arguments',
    'match_kwarg',
]


def _check_hamcrest_matcher(matcher: Matcher, item: Any) -> bool:
    # Add mismatch description
    if matcher.matches(item):
        return True
    else:
        # call mismatch description
        # TODO log mismatch
        return False


def match_arguments(expected_args: List[Any], actual_args: List[Any]) -> bool:
    expected_args_length = len(expected_args)
    actual_args_length = len(actual_args)
    if expected_args_length != actual_args_length:
        # TODO log mismatch
        return False

    zero_args = expected_args_length == 0 and actual_args_length == 0
    match = True if zero_args else False

    for index, arg_to_match in enumerate(expected_args):
        arg = actual_args[index]
        if isinstance(arg_to_match, Matcher):
            match = _check_hamcrest_matcher(arg_to_match, arg)
            if match:
                continue
            else:
                break
        else:
            match = arg_to_match == arg
            if match:
                continue
            else:
                break

    return match


def match_kwarg(expected_kwargs: Dict[str, Any], actual_kwars: Dict[str, Any]):
    expected_items = expected_kwargs.items()
    actual_items = actual_kwars.items()

    if len(expected_items) != len(actual_items):
        # TODO log mismatch
        return False

    for expected_key, expected_value in expected_items:
        matched = False
        for actual_key, actual_value in actual_items:
            matched_expected_key = expected_key == actual_key

            if matched_expected_key:
                matched = match_arguments([expected_value], [actual_value])
            if matched:
                # TODO consider removing the matched item
                break

        if not matched:
            # TODO log mismatch
            return False

    return True


